from django.contrib import admin
from . import models
from clouds.base.admin import StaticModelAdmin, OwnershipModelAdmin, OperatableAdminMixin, OperationAdmin
from django.utils.html import format_html
from django.urls import reverse
from django import forms
from django.db.models import Q
from dal import autocomplete
from user.utils import get_current_user

@admin.register(models.DataEngine)
class DataEngineAdmin(StaticModelAdmin):
    search_fields = ('description',)+StaticModelAdmin.search_fields
    class DataEngineForm(forms.ModelForm):
        class Meta:
            model = models.DataEngine
            fields = ('__all__')
            widgets = {
                'component': autocomplete.ModelSelect2(
                    url='dataenginecomponent-autocomplete',
                    forward=['engine']
                ),
            }
    form = DataEngineForm

@admin.register(models.DataSource)
class DataSourceAdmin(StaticModelAdmin):
    search_fields = ('description',)+StaticModelAdmin.search_fields
        
@admin.register(models.Dataset)
class DatasetAdmin(StaticModelAdmin):
    search_fields = ('description',)+StaticModelAdmin.search_fields
    def action(self,obj):
        return format_html('<a href="{}?dataset={}" class="button">Load</a>'.format(reverse('admin:data_datainstance_add'),obj.pk))
    extra=('action',)

@admin.register(models.DataInstance)
class DataInstanceAdmin(OwnershipModelAdmin,OperatableAdminMixin):
    class DataInstanceForm(forms.ModelForm):
        class Meta:
            model = models.DataInstance
            fields = ('__all__')
            widgets = {
                'engine': autocomplete.ModelSelect2(
                    url='datainstanceengine-autocomplete',
                    forward=['dataset','cluster']
                ),
            }
    form = DataInstanceForm
    def query(self, obj):
        # if not obj.ready: return None
        return format_html('<a href="{}" target="_blank" class="button">Query</a>'.format('obj.uri_elected'))
    # def action(self, obj):
    #     if obj.deleting:
    #         if not get_current_user().is_superuser: 
    #             return 'deleting'
    #     op_url=reverse('datainstanceoperation-list')
        # return self.action_button(obj,op_url)
    extra=('uri','query')#,'action'
    search_fields = ('name', 'dataset__name', 'cluster__name', 'engine__name')+OwnershipModelAdmin.search_fields
    list_filter = (
        ('dataset', admin.RelatedOnlyFieldListFilter),
        ('engine', admin.RelatedOnlyFieldListFilter),
        ('cluster', admin.RelatedOnlyFieldListFilter),
    )+OwnershipModelAdmin.list_filter
    def get_list_display_exclude(self, request, obj=None):
        if request.user.is_superuser: 
            return ()
        return ('owner','deleting')
    
@admin.register(models.DataInstanceOperation)
class DataInstanceOperationAdmin(OperationAdmin):
    def get_list_display(self,request,obj=None):
        return super().get_list_display(request,obj)+('log',)


from threading import local

_user = local()

class CurrentUserMiddleware(object):
    def __init__(self, get_response):
        self.get_response = get_response
    def __call__(self, request):
        _user.value = request.user
        return self.get_response(request)
def get_current_user():
    return _user.value


import paramiko
from scp import SCPClient
from django.conf import settings
import json
from io import StringIO
import calendar
import time
from django.urls import reverse
from django.utils.html import format_html

def get_url(obj):
    label= obj.name if getattr(obj,'name',False) else str(obj)
    url=reverse('admin:{}_{}_change'.format(obj._meta.app_label,obj._meta.model_name),args=(obj.pk,))
    return '<a href='+url+'>'+label+'</a>'

def get_formated_url(obj):
    return format_html(get_url(obj))

def get_current_timestamp():
    ts = calendar.timegm(time.gmtime())
    return ts

def get_refer_GET_parameter(request,name=None):
    from urllib import parse
    params=parse.parse_qs(parse.urlparse(request.META['HTTP_REFERER']).query)
    if not name:
        return params
    if name in params:
        return params[name][0]
    return None

def remedy_script_tidy(script, supervisor_operations):
    stats=[s.rstrip() for s in script.split('\n')]
    begin_i=0
    current_i=0
    length=len(stats)
    operations=[]
    while current_i < length:
        if not stats[current_i] or stats[current_i] in supervisor_operations:
            if current_i>begin_i:
                operations.append('\n'.join(stats[begin_i:current_i]))
            if stats[current_i]:
                operations.append(stats[current_i])  
            begin_i=current_i+1
        elif current_i==length-1:
            operations.append('\n'.join(stats[begin_i:current_i+1]))
        current_i+=1
    return operations

def remedy_script_hosts_add(hosts,overwrite=False):
    if overwrite: return "echo '{hosts}' >/etc/hosts".format(hosts=hosts)
    return "echo '{hosts}' >>/etc/hosts".format(hosts=hosts)

#TODO only support linux
def remedy_script_hosts_remove(hosts):
    return ';'.join(["sed -i '/{}/d' /etc/hosts".format(h) for h in hosts.split('\n') if h])
#TODO only support linux
def remedy_script_hosts_remove_from(tag):
    return "sed -i '/{}/,$d' /etc/hosts".format(tag)
#TODO only support linux
def remedy_script_hostname(hostname):
    return "hostnamectl set-hostname {hostname}".format(hostname=hostname)

#TODO make filesystem type chooseable
#must use rsync -ax to keep permissions
def remedy_script_mount_add(mount):
    return "mkfs.xfs {mount.dev}>/dev/null 2>&1 && " \
    "rsync -ax {mount.point} {mount.point}.old>/dev/null 2>&1 && " \
    "mkdir -p {mount.point} && " \
    "mount {mount.dev} {mount.point} && " \
    "rsync -ax {mount.point}.old/* {mount.point}>/dev/null 2>&1 && " \
    "rm -rf {mount.point}.old/ && " \
    "echo '{mount.dev} {mount.point} xfs defaults 0 2'>>/etc/fstab". \
    format(mount=mount)

def remedy_script_mount_remove(mount):
    return "sed -i '/{}/d' /etc/fstab".format('{mount.dev} {mount.point}'.format(mount=mount).replace('/', '\/'))

def remedy_image_ambari_agent():
    return 'curl -Ssl https://public-repo-1.hortonworks.com/ambari/centos7/2.x/updates/2.7.3.0/ambari.repo -o /etc/yum.repos.d/ambari.repo\n\nyum -q -y install ambari-agent'

def remedy_image_ambari_server():
    return 'yum -q -y install ambari-server\n\n' \
        'ambari-server setup -s \n\n' \
        'ambari-server start'

class SSH:
    def __init__(self,host,credential):
        self.client = paramiko.SSHClient()
        self.client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        credential=credential
        if 'password' not in credential:
            credential['password'] = None
        if 'port' not in credential:
            credential['port'] = 22
        if 'private_key' in credential:
            private_key_file=StringIO(credential['private_key'])
            credential['pkey']=paramiko.RSAKey.from_private_key(private_key_file)
            private_key_file.close()
        else:
            credential['pkey']=None
        mustend = time.time() + 600
        e=Exception()
        while time.time() < mustend:
            time.sleep(5)
            try:
                self.client.connect(
                    host,
                    username=credential['username'],
                    password=credential['password'],
                    pkey=credential['pkey'],
                    port=credential['port'],
                    timeout=None
                )
                # print("EOFError")
            except Exception as ex:
                e=ex
            else:
                return
        raise e

    def exec_batch(self, cmd):
        try:
            ftp = self.client.open_sftp()
            file=ftp.file('/tmp/packone.bash', "w", -1)
            file.write(cmd)
            file.flush()
            ftp.close()
            stdin, stdout, stderr = self.client.exec_command('sudo -uroot bash /tmp/packone.bash')
        except paramiko.SSHException as e:
            err='EXCEPTION MESSAGE:\n'+str(e)
            out=''
        else:
            err='\n'.join(stderr.readlines())
            if err: err='STDERR:\n'+err
            out='STDOUT:\n'+'\n'.join(stdout.readlines())
            if out.find('ERROR: Exiting with exit code 1')!=-1:
                err+=out
                out=''
        finally:
            return out, err

    def close(self):
        self.client.close()


from novaclient import client as nova_client
from novaclient.exceptions import NotFound as NovaNotFound
from cinderclient.exceptions import NotFound as CinderNotFound
from cinderclient import client as cinder_client
from uuid import uuid4
import time
from ..models import INSTANCE_STATUS

class Driver(object):
    def __init__(self, credential):
        self._credential=credential
        self._nova_client=nova_client.Client(credential['api_version'], username=credential['username'], password=credential['password'], project_name=credential['project_name'], auth_url=credential['auth_url'])
        self._cinder_client=cinder_client.Client(credential['api_version'],credential['username'],credential['password'],credential['project_name'],auth_url=credential['auth_url'])    
        self.instances=InstanceManager(self)
        self.volumes=VolumeManager(self)
        self.images=self._nova_client.glance
        self.flavors=self._nova_client.flavors

class InstanceManager(object):
    def __init__(self, driver):
        self.driver=driver
        self._manager=driver._nova_client.servers
        self.get=self._manager.get
        self.list=self._manager.list
        self.mountable_status=[INSTANCE_STATUS.active.value,INSTANCE_STATUS.shutdown.value]
    def create(self, image_id, template_id, remark=''):
        ins=self._manager.create(
            name=remark,
            image=image_id,
            flavor=template_id,
            security_groups=[self.driver._credential['security_group']],
            nics=[{'net-id':self.driver._credential['net-id']}],
            key_name=self.driver._credential['key_name'],
            # userdata="#cloud-config\n" \
            #         "ssh_pwauth: true\n" \
            #         "chpasswd:\n" \
            #         "  list: |\n" \
            #         "     root:bigdata\n" \
            #         # "     cloud-user:packone\n" \
            #         "     centos:bigdata\n" \
            #         "  expire: False\n",
            # files={
            #     '/etc/ssh/sshd_config':  "HostKey /etc/ssh/ssh_host_rsa_key\n" \
            #                             "HostKey /etc/ssh/ssh_host_ecdsa_key\n" \
            #                             "SyslogFacility AUTHPRIV\n" \
            #                             "PermitRootLogin yes\n" \
            #                             "AuthorizedKeysFile	.ssh/authorized_keys\n" \
            #                             "PasswordAuthentication yes\n" \
            #                             "ChallengeResponseAuthentication no\n" \
            #                             "GSSAPIAuthentication yes\n" \
            #                             "GSSAPICleanupCredentials yes\n" \
            #                             "UsePAM yes\n" \
            #                             "X11Forwarding yes\n" \
            #                             "UsePrivilegeSeparation sandbox		# Default for new installations.\n" \
            #                             "AcceptEnv LANG LC_CTYPE LC_NUMERIC LC_TIME LC_COLLATE LC_MONETARY LC_MESSAGES\n" \
            #                             "AcceptEnv LC_PAPER LC_NAME LC_ADDRESS LC_TELEPHONE LC_MEASUREMENT\n" \
            #                             "AcceptEnv LC_IDENTIFICATION LC_ALL LANGUAGE\n" \
            #                             "AcceptEnv XMODIFIERS\n" \
            #                             "Subsystem	sftp	/usr/libexec/openssh/sftp-server"
                
            # }
        )
        mustend = time.time() + 600
        while time.time() < mustend:
            ins=self.get(ins.id)
            if 'provider' in ins.addresses:
                break
            time.sleep(5)
        return ins
    def delete(self, instance_id):
        try:
            return self._manager.delete(instance_id)
        except NovaNotFound as e:
            print(e)
    def force_delete(self, instance_id):
        return self.delete(instance_id)
    def get_status(self, instance_id):
        ins=self.get(instance_id)
        if ins.status=='ACTIVE':
            return INSTANCE_STATUS.active.value
        elif ins.status=='PAUSED':
            return INSTANCE_STATUS.pause.value
        elif ins.status=='BUILDING':
            return INSTANCE_STATUS.preparing.value
        elif ins.status=='STOPPED':
            return INSTANCE_STATUS.shutdown.value
        elif ins.status=='SHUTOFF':
            return INSTANCE_STATUS.poweroff.value
        elif ins.status=='ERROR':
            return INSTANCE_STATUS.failure.value
        return INSTANCE_STATUS.null.value

class VolumeManager(object):
    def __init__(self, driver):
        self.driver=driver
        self._manager=driver._cinder_client.volumes
        self.get=self._manager.get
        self.list=self._manager.list
    def create(self, size, remark=''):
        volume=self._manager.create(
            name=remark,
            size=size
        )
        mustend = time.time() + 60
        while time.time() < mustend:
            volume=self.get(volume.id)
            if volume.status == 'available': break
            time.sleep(5)
        return volume
    def delete(self, volume_id):
        mustend = time.time() + 60
        while time.time() < mustend:
            try:
                volume=self._manager.get(volume_id)
            except CinderNotFound as e:
                print(e)
                return
            if volume.status == 'available': break
            time.sleep(5)
        return volume.delete()
    def mount(self, volume_id, instance_id):
        self.driver._nova_client.volumes.create_server_volume(server_id=instance_id, volume_id=volume_id)
        mustend = time.time() + 60
        while time.time() < mustend:
            volume=self.get(volume_id)
            if volume.status == 'in-use': break
            time.sleep(5)
        return volume
    def unmount(self, volume_id, instance_id):
        mustend = time.time() + 60
        while time.time() < mustend:
            ins=self.driver.instances.get(instance_id)
            if ins.status == 'SHUTOFF': break
            time.sleep(5)
        return self.driver._nova_client.volumes.delete_server_volume(server_id=instance_id, volume_id=volume_id)


from django.contrib import admin
from . import models
from clouds.base.admin import StaticModelAdmin, OwnershipModelAdmin, OperationAdmin, OperatableAdminMixin, AutoModelAdmin, M2MOperationAdmin
from clouds.models import Cloud, Group, InstanceBlueprint, InstanceOperation, GroupOperation
from user.models import Balance
from django.utils.html import format_html
from django.urls import reverse
from django.db.models import Q
from django import forms
from dal import autocomplete
from user.utils import get_current_user
from clouds.utils import get_url, get_formated_url

@admin.register(models.Stack)
class StackAdmin(StaticModelAdmin):
    list_filter = ('_driver',)+StaticModelAdmin.list_filter
    def import_engine(modeladmin, request, queryset):
        for stack in queryset:
            stack.import_engine()
    import_engine.short_description = "Refresh engines from selected stacks"
    actions = [import_engine]

@admin.register(models.Engine)
class EngineAdmin(StaticModelAdmin):
    pass

@admin.register(models.Scale)
class ScaleAdmin(StaticModelAdmin):
    list_filter = ('auto',)+StaticModelAdmin.list_filter
    def get_queryset_Q(self, request):
        return Q(pk__in=request.user.scales())

@admin.register(models.Cluster)
class ClusterAdmin(OwnershipModelAdmin,OperatableAdminMixin):
    def access(self, obj):
        if not obj.ready: return None
        return format_html('<a href="{}" target="_blank" class="button">Manage</a>'.format(obj.portal))
    def instances(self,object):
        return format_html('<br/>'.join([get_url(ins) for ins in object.get_instances()]))  
    def action(self, obj):
        if obj.deleting:
            if not get_current_user().is_superuser:
                return 'deleting'
        op_url=reverse('clusteroperation-list')
        return self.action_button(obj,op_url)
    search_fields = ('name','scale__name')+OwnershipModelAdmin.search_fields
    list_filter = (('scale', admin.RelatedOnlyFieldListFilter),)+OwnershipModelAdmin.list_filter
    extra=('access','action','instances')
    def get_list_display_exclude(self, request, obj=None):
        if request.user.is_superuser: return ('instances',)
        return ('owner','deleting','instances')
    def start(modeladmin, request, queryset):
        for cluster in queryset:
            models.ClusterOperation(
                target=cluster,
                operation=models.INSTANCE_OPERATION.start.value,
                status=models.OPERATION_STATUS.running.value
            ).save()
    start.short_description = "Start selected clusters"
    def materialize(modeladmin, request, queryset):
        for cluster in queryset:
            if not cluster.name.startswith('bootstrap.'): continue
            for ins in cluster.get_instances():
                try:
                    ins.cloud.driver.instances.get(str(ins.uuid)).create_image(ins.image.name.replace('-bootstrap',''))
                except Exception as e:
                    print(e)
            ins.cloud.import_image()
            image_master1=ins.cloud.image_set.get(name='packone-master1')
            image_master2=ins.cloud.image_set.get(name='packone-master2')
            image_slave=ins.cloud.image_set.get(name='packone-slave')
            blueprint_master1, created=ins.cloud.instanceblueprint_set.get_or_create(
                name='packone-master1',
                cloud=ins.cloud,
                template=ins.template,
                image=image_master1,
                volume_capacity=500,
                remark='auto created',
                owner=ins.owner
            )
            blueprint_master2, created=ins.cloud.instanceblueprint_set.get_or_create(
                name='packone-master2',
                cloud=ins.cloud,
                template=ins.template,
                image=image_master2,
                volume_capacity=500,
                remark='auto created',
                owner=ins.owner
            )
            blueprint_slave, created=ins.cloud.instanceblueprint_set.get_or_create(
                name='packone-slave',
                cloud=ins.cloud,
                template=ins.template,
                image=image_slave,
                volume_capacity=500,
                remark='auto created',
                owner=ins.owner
            )
            from .utils import remedy_scale_ambari_fast_init, remedy_scale_ambari_fast_scale_out, remedy_scale_ambari_fast_scale_in
            s, created=models.Scale.objects.get_or_create(
                name='packone.{}'.format(ins.cloud.name),
                _remedy_script=remedy_scale_ambari_fast_init(),
                _remedy_script_scale_out=remedy_scale_ambari_fast_scale_out(),
                _remedy_script_scale_in=remedy_scale_ambari_fast_scale_in(),
                owner=ins.owner
            )
            if created:
                s.init_blueprints.add(*(blueprint_master1, blueprint_master2, blueprint_slave))
                s.step_blueprints.add(*(blueprint_slave,))
    materialize.short_description = "Materialize the cluster as a scale"
    def scale_out(modeladmin, request, queryset):
        for cluster in queryset:
            cluster.scale_one_step()
    scale_out.short_description = "Scale out one step"
    def scale_in(modeladmin, request, queryset):
        for cluster in queryset:
            cluster.steps.last().delete()
    scale_in.short_description = "Scale in one step"
    def destroy(modeladmin, request, queryset):
        for cluster in queryset:
            cluster.delete()
    destroy.short_description = "Destroy selected clusters"
    actions=[start,materialize,scale_out,scale_in,destroy]
    def has_delete_permission(self, request, obj=None):
        return False
    def get_form_field_queryset_Q(self, db_field, request):
        if db_field.name == 'scale': return Q(pk__in=request.user.scales())
        return None
    def get_queryset_Q(self, request):
        return Q(pk__in=request.user.clusters())
    
@admin.register(models.ClusterOperation)
class ClusterOperationAdmin(M2MOperationAdmin):
    def has_module_permission(self, request):
        return False

@admin.register(models.StepOperation)
class StepOperationAdmin(M2MOperationAdmin):
    def _target(self,obj):
        return  format_html(get_url(obj.cluster)+'/'+get_url(obj.target))
    def get_queryset_Q(self, request):
        return Q(target__in=request.user.steps()) & ~Q(status=models.OPERATION_STATUS.success.value)
    def has_add_permission(self, request, obj=None):
        return False
    def has_change_permission(self, request, obj=None):
        return False
    def has_view_permission(self, request, obj=None):
        return True
    def has_delete_permission(self, request, obj=None):
        return True
    def has_module_permission(self, request):
        return True



def remedy_scale_ambari_bootstrap():
    return "sed -i 's/hostname=localhost/hostname=master1.packone/g' /etc/ambari-agent/conf/ambari-agent.ini\n\n" \
        "ambari-agent start >/dev/null 2>&1\n\n" \
        'if [ `hostname` == "master1.packone" ]; then\n' \
        'sleep 10\n' \
        'yum -q -y install epel-release\n' \
        'yum -q -y install python-pip\n' \
        'pip --disable-pip-version-check install ambari\n' \
        'ambari localhost:8080 cluster create packone typical_triple master1.packone master2.packone slave.packone\n' \
        "fi"

def remedy_scale_ambari_fast_init():
    return 'rm -rf /hadoop\n' \
        'mkdir -p /data/hadoop\n' \
        'ln -sf /data/hadoop /hadoop\n' \
        '#env\n' \
        "echo 'JAVA_HOME=/usr/jdk64/default'>>/etc/profile.d/packone-java.sh\n" \
        "echo 'JRE_HOME=/usr/jdk64/default/jre'>>/etc/profile.d/packone-java.sh\n" \
        "echo 'CLASS_PATH=.:$JAVA_HOME/lib:$JRE_HOME/lib'>>/etc/profile.d/packone-java.sh\n" \
        "echo 'PATH=$JAVA_HOME/bin:$JRE_HOME/bin:$PATH'>>/etc/profile.d/packone-java.sh\n" \
        'reboot\n' \
        'if [ `hostname` == "master1.packone" ]; then\n' \
        '    sleep 60\n' \
        '    pip --disable-pip-version-check install pk1-remedy\n' \
        '    ambari-service-start master1.packone:8080\n' \
        'fi\n' \

def remedy_scale_ambari_fast_scale_out():
    return "rm -rf /hadoop\n" \
        "mkdir -p /data/hadoop\n" \
        "ln -sf /data/hadoop /hadoop\n" \
        "#env\n" \
        "echo 'JAVA_HOME=/usr/jdk64/default'>>/etc/profile.d/packone-java.sh\n" \
        "echo 'JRE_HOME=/usr/jdk64/default/jre'>>/etc/profile.d/packone-java.sh\n" \
        "echo 'CLASS_PATH=.:$JAVA_HOME/lib:$JRE_HOME/lib'>>/etc/profile.d/packone-java.sh\n" \
        "echo 'PATH=$JAVA_HOME/bin:$JRE_HOME/bin:$PATH'>>/etc/profile.d/packone-java.sh\n" \
        "reboot\n" \
        "pip --disable-pip-version-check install pk1-remedy\n" \
        "ambari-host-clone master1.packone:8080 slave.packone `hostname`\n" \

def remedy_scale_ambari_fast_scale_in():
    return '#ambari-host-delete master1.packone:8080 `hostname`'